;;;
;;; Solves the coin change problem[0], taking into account a given finite limit
;;; of coins of each denomination, using a "dumb" greedy algorithm. This
;;; doesn't necessarily result in the minimum number of coins for non-canonical
;;; sets of denominations, but is in practice not far from it, and is simple to
;;; implement. Uses a couple of simple checks to fail early in certain
;;; scenarios, and a memoization table, from target to denomination to solution
;;; (this works because of the greedy nature). The recursive level will be at
;;; most the number of denominations.
;;;
;;; Mathematically speaking, given target T, denominations di, and counts ci,
;;; it computes the coefficients si of the following linear equation, such that
;;; 0 <= si <= ci:
;;;   T = (sum (* si di))
;;;
;;; This is NOT the solutions counting algorithm!
;;;
;;; [0]: https://en.wikipedia.org/wiki/Change-making_problem
;;;

(module coin-change
  (coin-change)

  (import
    scheme

    (only (chicken base)
          alist-ref
          alist-update
          define-constant
          foldl
          gensym
          print
          receive
          sub1
          )
    (only (chicken sort) sort)
    (chicken type)

    (only (srfi 1) any)
    )

  (define-constant missing 'missing)
  (define (missing? obj) (eq? obj missing))

  ; TODO: better performant impl
  (define (alist-update-with key alist update #!optional (=? eq?))
    (let* ((value (alist-ref key alist =? missing))
           (value (if (missing? value) (update) (update value))))
      (alist-update key value alist =?)))

  ; Memoization table (solution may be #f):
  ; `((,target . ((,denomination . ,solution)
  ;               ...))
  ;   (,target . ((,denomination . ,solution)
  ;               ...))
  ;   ...)
  (define (memo:get memo target denomination)
    (alist-ref denomination (alist-ref target memo = '()) = missing))

  (define (memo:set memo target denomination solution)
    (alist-update-with target memo
                       (lambda (#!optional (denoms-alist '()))
                         (alist-update denomination denoms-alist solution =))
                       =))

  ; `(,denom ,count . ,csum)
  (define get-denom car)
  (define get-count cadr)
  (define get-csum cddr)

  (define (max-denomination-multiplicity denomination count target)
    (min count (inexact->exact (floor (/ target denomination)))))

  ; TODO: memoize this too?
  (define (has-divisor? denominations target)
    (define (divides? denomination)
      (zero? (modulo target (get-denom denomination))))
    (any divides? denominations))

  ; Sorts the input denominations list in ascending order, and includes a
  ; cumulative sum of the value at each "level".
  ;
  ;   (prepare-denominations '((3 . 3) (2 . 2) (5 . 5)))
  ;   => '((5 5 . 38)
  ;        (3 3 . 13)
  ;        (2 2 . 4))
  (define (prepare-denominations denominations)
    (define (</denom d1 d2) (< (car d1) (cdr d2)))
    (define (kons sum.denominations denomination)
      (let ((sum (car sum.denominations))
            (denominations (cdr sum.denominations))
            (denom (car denomination))
            (count (cdr denomination)))
        (let ((sum (+ sum (* count denom))))
          (cons sum
                `((,denom ,count . ,sum) . ,denominations)))))
    (let* ((denominations (sort denominations </denom))
           (sum.denominations (foldl kons '(0 . ()) denominations)))
      (cdr sum.denominations)))

  ; target :: Int
  ; denominations :: [(Denomination . Count)]
  ;
  ;   (coin-change 765 '((50 . 123) (10 . 456) (5 . 789)))
  ;   => (values ((50 . 15) (10 . 1) (5 . 1))
  ;              ((5 (5)) (15 (10) (5 . 1)) (765 (50) (10 . 1) (5 . 1))))
  (: coin-change ((list-of (pair integer integer)) integer --> (or false (list-of (pair integer integer))) list))
  (define (coin-change target denominations)
    ; Internal entry-point
    (define (coin-change* memo target denominations)
      (cond
        ((zero? target)
         (values '() memo))
        ; There is no solution in any of these cases:
        ;  * There are no coins
        ;  * The target is greater than the cumulative sum
        ;  * The target is not divisible by any of the denominations
        ((or (null? denominations)
             (> target (get-csum (car denominations)))
             (not (has-divisor? denominations target)))
         (values #f memo))
        (else
          (let ((denomination (get-denom (car denominations)))
                (count (get-count (car denominations)))
                (denominations (cdr denominations)))
            (let ((solution (memo:get memo target denomination)))
              (if (missing? solution)
                  (receive (solution memo) (coin-change/multiplicity memo denomination count denominations target
                                                                     (max-denomination-multiplicity denomination count target))
                    (values solution (memo:set memo target denomination solution)))
                  (values solution memo)))))))

      ; Tests all multiplicities of a given denomination, from greatest to least
      (define (coin-change/multiplicity memo denomination count denominations target nd)
        (if (negative? nd)
            (values #f memo) ; No solution found for all possible multiplicities of this denomination
            (receive (solution memo) (coin-change* memo (- target (* nd denomination)) denominations)
              (if solution ; Return the first solution found with this multiplicity
                  (values `((,denomination . ,nd) . ,solution) memo)
                  ; Try a lower multiplicity of this denomination
                  (coin-change/multiplicity memo denomination count denominations target (sub1 nd))))))

      (coin-change* '() target (prepare-denominations denominations)))
    )
